#include <Arduino.h>
#include "HS_AsyncTCP_base.h"
#include "BufferStream.h"
#include "StreamBuffers.h"
#include "millisDelay.h"  // from SafeString library

// user code must implement these methods
extern void asyncSetup();
extern void asyncConnected(Stream &stream);
extern void asyncDisconnected();
extern void asyncDataReceived(Stream &stream);
extern void asyncLoop(Stream &stream);

static millisDelay connectionTimeout;
static unsigned long connection_timeout_ms = 0;
static void connect_cb(void* arg, HS_AsyncClient* newClientPtr);
static Stream *wifiDebugPtr;
static bool sendResponse = false; // true if the outBufferWiFi holds a cmd response

static HS_AsyncServer *serverPtr;
static HS_AsyncClient* clientPtr; // set to NULL on disconnect
static volatile bool disconnectNow = false;

static BufferStream inBuffer;
static BufferStream outBuffer;

static StreamBuffers streamBuffers(inBuffer, outBuffer);

static BufferStream inBufferWiFi;
static BufferStream outBufferWiFi;
static StreamBuffers streamBuffersWiFi(inBuffer, outBuffer);
static BufferStream secondOutBufferWiFi;

static millisDelay sendTimer;
static unsigned long SEND_TIMER_MS = 250;

// call this to set connection timeout if sending keepalives
// else leave as 0 i.e. not connection timeout
void asyncConnectionTimeout(unsigned long ms_timeout) {
  connection_timeout_ms = ms_timeout;
}

void setAsyncDebugPtr(Stream *debugPtr) {
  wifiDebugPtr = debugPtr;
}

bool initAsyncServer(uint16_t port) {
  if (serverPtr) {
    return true; // already done
  }
  serverPtr = new HS_AsyncServer(port);
  if (!serverPtr) {
    return false;
  }
  serverPtr->setNoDelay(true);
  serverPtr->onClient(connect_cb, 0);
  sendTimer.start(SEND_TIMER_MS);
  return serverPtr->begin();
}

void asyncCloseConnection() {
  disconnectNow = true;
}

static void stopClient() {
  streamBuffers.clear(); // clean up
  if (clientPtr) {
    clientPtr->stop();
    clientPtr = NULL;
  }
  disconnectNow = false;
}

//#define DEBUG_CMD_WRITE
//to print cmd response handling to Serial
#ifdef DEBUG_CMD_WRITE
static bool printR = false;
#endif

static void sendOutBuffer() { // sends outBufferWiFi if can else skip it, clearsOutBufferWiFi
  if (clientPtr && (clientPtr->connected()) && (clientPtr->canSend())) {
    if (outBufferWiFi.available()) {
      if (wifiDebugPtr) {
        wifiDebugPtr->print("sending "); wifiDebugPtr->println(outBufferWiFi.available());
      }
      clientPtr->write(outBufferWiFi.getBuffer(), outBufferWiFi.available());
      if (sendResponse) {
#ifdef DEBUG_CMD_WRITE            
        if (printR) {
          Serial.println("-R"); // flag cmd response that was held now sent
         Serial.write((uint8_t*)outBufferWiFi.getBuffer(), outBufferWiFi.available());Serial.println();
        }
        printR = false;
#endif         
        sendResponse = false; 
      }
      outBufferWiFi.clear();
    }
  } else {
    if (wifiDebugPtr) {
      wifiDebugPtr->println("skip write as not connected or waiting for Ack");
    }
     if (!sendResponse) {
      outBufferWiFi.clear();
    } else {  // skipped sending cmd response,hold to send later
#ifdef DEBUG_CMD_WRITE            
      if (!printR) {
        printR = true;
        Serial.println("R"); // flag cmd response being held
      }
#endif         
    }
  }
}

static void copyData(bool cmdResponse) { // copies outBuffer to outBufferWiFi is there is space, else sendOutBuffer
  if (outBuffer.available()) {
  } else {
      return;
  }
  if (outBuffer.available() > outBufferWiFi.availableForWrite()) { // no space in outBufferWiFi for all of the data
      sendOutBuffer(); // if not sendResponse set, then this ALWAYS clears the outBufferWiFi
  }
  if (cmdResponse) {
      sendResponse = true;
  }
  if (outBuffer.available() < outBufferWiFi.availableForWrite()) {
    outBufferWiFi.write((uint8_t*)outBuffer.getBuffer(), outBuffer.available());
    outBuffer.clear();
  } // else keep outBuffer for later send
}

static void loopHandler(void* arg, HS_AsyncClient* newClientPtr) {
  if (connectionTimeout.justFinished() || disconnectNow) {
    stopClient();
    return;
  }
  if ((!clientPtr) || (!clientPtr->connected())) {
      outBuffer.clear();
      outBufferWiFi.clear();
      return;
  }
  if (sendTimer.justFinished()) {
      sendOutBuffer();
      sendTimer.restart();
  }   
  
  asyncLoop(streamBuffers);
  if (!sendResponse) { // still trying to send cmd response
    copyData(false); // copy and send or skip, clears outBuffer
  } else {
    outBuffer.clear();
    sendOutBuffer(); 
  }
  
  if (outBufferWiFi.available() &&  (!sendTimer.isRunning())) {
      sendTimer.restart();
  }
}

// not used
static void pollHandler(void* arg, HS_AsyncClient* newClientPtr) {
  if (wifiDebugPtr) {
//    wifiDebugPtr->print("got poll"); wifiDebugPtr->println();
  }
}

static void ackHandler(void* arg, HS_AsyncClient* clientPtr, size_t len, uint32_t time) {
  if (wifiDebugPtr) {
    wifiDebugPtr->print("got ack len:"); wifiDebugPtr->print(len); wifiDebugPtr->print(" time:"); wifiDebugPtr->println(time);
  }
  if (connection_timeout_ms) {
    connectionTimeout.restart();
  }
}

static void dataRecieved_cb(void*arg, HS_AsyncClient*client, void *data, size_t len) {
  if (wifiDebugPtr) {
    wifiDebugPtr->print("dataRecieved_cb core:"); wifiDebugPtr->println(xPortGetCoreID());
  }
  if (connection_timeout_ms) {
    connectionTimeout.restart();
  }
  sendOutBuffer(); // clears outBufferWiFi
  copyData(false); // clears outBuffer
  
  inBuffer.clear();
  inBuffer.write((uint8_t*)data, len);
  if (wifiDebugPtr) {
    wifiDebugPtr->println("got data");
    wifiDebugPtr->println(inBuffer.getBuffer());
  }
  asyncDataReceived(streamBuffers); // may write response msg
  if (wifiDebugPtr) {
    wifiDebugPtr->println(outBuffer.getBuffer());
  }
  if (outBuffer.available()) {
    if (!sendTimer.isRunning()) {
      sendTimer.restart();
    }
  }
  copyData(true); // clears outBuffer, outBufferWiFi cleared above

}

static void disconnect_cb(void* arg, HS_AsyncClient* existingClientPtr) {
  // have disconnected so clean up this client
  if (wifiDebugPtr) {
    wifiDebugPtr->println("disconnected");
  }
  connectionTimeout.stop();
  stopClient(); // cleans up
  delete existingClientPtr;
  asyncDisconnected(); // inform user
}

static void connect_cb(void* arg, HS_AsyncClient* newClientPtr) {
  if (wifiDebugPtr) {
    wifiDebugPtr->println("got server connect");
  }
  if (connection_timeout_ms) {
    connectionTimeout.start(connection_timeout_ms);
  }
  clientPtr = newClientPtr;
  clientPtr->onData(dataRecieved_cb);
  clientPtr->onDisconnect(disconnect_cb);
  clientPtr->onAck(ackHandler);
  //  clientPtr->onPoll(pollHandler); // not used
  clientPtr->onLoop(loopHandler);
  streamBuffers.clear();
  asyncConnected(streamBuffers);
}
